from enum import Enum

DEFAULT_KW_ROUND_PERSICION = 1


class User(Enum):
    INSTALLER = "installer"
    CUSTOMER = "customer"
    ENGINEER = "engineer"
    KIOSK = "kiosk"
    ADMIN = "admin"


class Roles(Enum):
    HOME_OWNER = "Home_Owner"
    KIOSK_VIEWER = "Kiosk_Viewer"
    PROVIDER_ENGINEER = "Provider_Engineer"
    TESLA_ENGINEER = "Tesla_Engineer"


class GridStatus(Enum):
    CONNECTED = "SystemGridConnected"
    ISLANDED_READY = "SystemIslandedReady"
    ISLANDED = "SystemIslandedActive"
    TRANSITION_TO_GRID = "SystemTransitionToGrid"  # Used in version 1.46.0
    TRANSITION_TO_ISLAND = "SystemTransitionToIsland"


class IslandMode(Enum):
    OFFGRID = "intentional_reconnect_failsafe"
    ONGRID = "backup"


class GridState(Enum):
    COMPLIANT = "Grid_Compliant"
    QUALIFINY = "Grid_Qualifying"
    UNCOMPLIANT = "Grid_Uncompliant"


class LineStatus(Enum):
    NON_BACKUP = "NonBackup"
    BACKUP = "Backup"
    NOT_CONFIGURED = "NotConfigured"


class OperationMode(Enum):
    BACKUP = "backup"
    SELF_CONSUMPTION = "self_consumption"
    AUTONOMOUS = "autonomous"
    SCHEDULER = "scheduler"
    SITE_CONTROL = "site_control"


SUPPORTED_OPERATION_MODES = [
    OperationMode.BACKUP,
    OperationMode.SELF_CONSUMPTION,
    OperationMode.AUTONOMOUS,
]


class InterfaceType(Enum):
    ETH = "EthType"
    GSM = "GsmType"
    WIFI = "WifiType"


class MeterType(Enum):
    SOLAR = "solar"
    SITE = "site"
    BATTERY = "battery"
    LOAD = "load"
    GENERATOR = "generator"
    BUSWAY = "busway"


class DeviceType(Enum):
    """
    Devicetype as returned by "device_type"
    GW1: Gateway 1
    GW2: Gateway 2
    SMC: ?
    """

    GW1 = "hec"
    GW2 = "teg"
    SMC = "smc"


class SyncType(Enum):
    V1 = "v1"
    V2 = "v2"
    V2_1 = "v2.1"


class UpdateState(Enum):
    CHECKING = "/clear_update_status"
    SUCCEEDED = "/update_succeeded"
    FAILED = "/update_failed"
    STAGED = "/update_staged"
    DOWNLOAD = "/download"
    DOWNLOADED = "/update_downloaded"
    UNKNOWN = "/update_unknown"


class UpdateStatus(Enum):
    IGNORING = "ignoring"
    ERROR = "error"
    NONACTIONABLE = "nonactionable"



from typing import List, Union

from .const import MeterType


class PowerwallError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class ApiError(PowerwallError):
    def __init__(self, error: str):
        super().__init__("Powerwall api error: {}".format(error))


class MissingAttributeError(ApiError):
    def __init__(self, response: dict, attribute: str, url: Union[str, None] = None):
        self.response: dict = response
        self.attribute: str = attribute
        self.url: Union[str, None] = url

        if url is None:
            super().__init__(
                "The attribute '{}' is expected in the response but is missing.".format(
                    attribute
                )
            )
        else:
            super().__init__(
                "The attribute '{}' is expected in the response for \
                 '{}' but is missing.".format(
                    attribute, url
                )
            )


class PowerwallUnreachableError(PowerwallError):
    def __init__(self, reason: Union[str, None] = None):
        msg = "Powerwall is unreachable"
        self.reason: Union[str, None] = reason
        if reason is not None:
            msg = "{}: {}".format(msg, reason)
        super().__init__(msg)


class AccessDeniedError(PowerwallError):
    def __init__(
        self,
        resource: str,
        error: Union[str, None] = None,
        message: Union[str, None] = None,
    ):
        self.resource: str = resource
        self.error: Union[str, None] = error
        self.message: Union[str, None] = message
        msg = "Access denied for resource {}".format(resource)
        if error is not None:
            if message is not None:
                msg = "{}: {}: {}".format(msg, error, message)
            else:
                msg = "{}: {}".format(msg, error)
        super().__init__(msg)


class MeterNotAvailableError(PowerwallError):
    def __init__(self, meter: MeterType, available_meters: List[MeterType]):
        self.meter: MeterType = meter
        self.available_meters: List[MeterType] = available_meters
        super().__init__(
            "Meter {} is not available at your powerwall. \
             Following meters are available: {} ".format(
                meter.value, available_meters
            )
        )



from http.client import responses
from json.decoder import JSONDecodeError
from typing import Any, List, Optional
from urllib.parse import urljoin

import requests
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning

from .error import AccessDeniedError, ApiError, PowerwallUnreachableError


class API(object):
    def __init__(
        self,
        endpoint: str,
        timeout: int = 10,
        http_session: Optional[requests.Session] = None,
        verify_ssl: bool = False,
        disable_insecure_warning: bool = True,
    ) -> None:
        if disable_insecure_warning:
            disable_warnings(InsecureRequestWarning)

        self._endpoint = self._parse_endpoint(endpoint)
        self._timeout = timeout
        self._http_session = http_session if http_session else requests.Session()
        self._http_session.verify = verify_ssl

    @staticmethod
    def _parse_endpoint(endpoint: str) -> str:
        if endpoint.startswith("https"):
            endpoint = endpoint
        elif endpoint.startswith("http"):
            endpoint = endpoint.replace("http", "https")
        else:
            # Use str.format instead of f'strings to be backwards compatible
            endpoint = "https://{}".format(endpoint)

        if not endpoint.endswith("api") and not endpoint.endswith("/"):
            endpoint += "/api/"
        elif endpoint.endswith("api"):
            endpoint += "/"
        elif endpoint.endswith("/") and not endpoint.endswith("api/"):
            endpoint += "api/"

        return endpoint

    @staticmethod
    def _handle_error(response: requests.Response) -> None:
        if response.status_code == 404:
            raise ApiError(
                "The url {} returned error 404".format(response.request.path_url)
            )

        if response.status_code == 401 or response.status_code == 403:
            response_json = None
            try:
                response_json = response.json()
            except Exception:
                raise AccessDeniedError(response.request.path_url)
            else:
                raise AccessDeniedError(
                    response.request.path_url,
                    response_json.get("error"),
                    response_json.get("message"),
                )

        if response.text is not None and len(response.text) > 0:
            raise ApiError(
                "API returned status code '{}: {}' with body: {}".format(
                    response.status_code,
                    responses.get(response.status_code),
                    response.text,
                )
            )
        else:
            raise ApiError(
                "API returned status code '{}: {}' ".format(
                    response.status_code, responses.get(response.status_code)
                )
            )

    def _process_response(self, response: requests.Response) -> dict:
        if response.status_code >= 400:
            # API returned some sort of error that must be handled
            self._handle_error(response)

        if len(response.content) == 0:
            return {}

        try:
            response_json = response.json()
        except JSONDecodeError:
            raise ApiError(
                "Error while decoding json of response: {}".format(response.text)
            )

        if response_json is None:
            return {}

        # Newer versions of the powerwall do not return such values anymore
        # Kept for backwards compability or if the API changes again
        if "error" in response_json:
            raise ApiError(response_json["error"])

        return response_json

    def url(self, path: str):
        return urljoin(self._endpoint, path)

    def get(self, path: str, headers: dict = {}) -> Any:
        try:
            response = self._http_session.get(
                url=self.url(path),
                timeout=self._timeout,
                headers=headers,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            raise PowerwallUnreachableError(str(e))

        return self._process_response(response)

    def post(
        self,
        path: str,
        payload: dict,
        headers: dict = {},
    ) -> Any:
        try:
            response = self._http_session.post(
                url=self.url(path),
                json=payload,
                timeout=self._timeout,
                headers=headers,
            )
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            raise PowerwallUnreachableError(str(e))

        return self._process_response(response)

    def is_authenticated(self) -> bool:
        return "AuthCookie" in self._http_session.cookies.keys()

    def login(
        self,
        username: str,
        email: str,
        password: str,
        force_sm_off: bool = False,
    ) -> dict:
        # force_sm_off is referred to as 'shouldForceLogin' in the web source code
        return self.post(
            "login/Basic",
            {
                "username": username,
                "email": email,
                "password": password,
                "force_sm_off": force_sm_off,
            },
        )

    def logout(self) -> None:
        if not self.is_authenticated():
            raise ApiError("Must be logged in to log out")
        # The api unsets the auth cookie and the token is invalidated
        self.get("logout")

    def close(self) -> None:
        # Close the HTTP Session
        # THis method is required for testing,
        # so python doesn't complain about unclosed resources
        self._http_session.close()

    # Endpoints are mapped to one method by <verb>_<path> so they can be easily accessed

    def get_system_status(self) -> dict:
        return self.get("system_status")

    def get_system_status_soe(self) -> dict:
        return self.get("system_status/soe")

    def get_meters_aggregates(self) -> dict:
        return self.get("meters/aggregates")

    def get_sitemaster_run(self):
        return self.get("sitemaster/run")

    def get_sitemaster_stop(self):
        return self.get("sitemaster/stop")

    def get_sitemaster(self) -> dict:
        return self.get("sitemaster")

    def get_status(self) -> dict:
        return self.get("status")

    def get_customer_registration(self) -> dict:
        return self.get("customer/registration")

    def get_powerwalls(self):
        return self.get("powerwalls")

    def get_operation(self) -> dict:
        return self.get("operation")

    def get_networks(self) -> list:
        return self.get("networks")

    def get_phase_usage(self):
        return self.get("powerwalls/phase_usages")

    def post_sitemaster_run_for_commissioning(self):
        return self.post("sitemaster/run_for_commissioning", payload={})

    def get_solars(self):
        return self.get("solars")

    def get_config(self):
        return self.get("config")

    def get_logs(self):
        return self.get("getlogs")

    def get_meters(self) -> list:
        return self.get("meters")

    def get_meters_site(self) -> list:
        return self.get("meters/site")

    def get_meters_solar(self) -> list:
        return self.get("meters/solar")

    def get_installer(self) -> dict:
        return self.get("installer")

    def get_solar_brands(self) -> List[str]:
        return self.get("solars/brands")

    def get_system_update_status(self) -> dict:
        return self.get("system/update/status")

    def get_system_status_grid_status(self) -> dict:
        return self.get("system_status/grid_status")

    def get_site_info(self) -> dict:
        return self.get("site_info")

    def get_site_info_grid_codes(self) -> list:
        return self.get("site_info/grid_codes")

    def post_site_info_site_name(self, body: dict) -> dict:
        return self.post("site_info/site_name", body)

    def post_islanding_mode(self, body: dict) -> dict:
        return self.post("v2/islanding/mode", body)



from typing import List, Union

import requests

from .api import API
from .const import DeviceType, GridStatus, IslandMode, OperationMode, User
from .error import ApiError
from .helpers import assert_attribute
from .responses import (
    BatteryResponse,
    LoginResponse,
    MeterDetailsResponse,
    MetersAggregatesResponse,
    PowerwallStatusResponse,
    SiteInfoResponse,
    SiteMasterResponse,
    SolarResponse,
)


class Powerwall:
    def __init__(
        self,
        endpoint: str,
        timeout: int = 10,
        http_session: Union[requests.Session, None] = None,
        verify_ssl: bool = False,
        disable_insecure_warning: bool = True,
    ):
        self._api = API(
            endpoint,
            timeout,
            http_session,
            verify_ssl,
            disable_insecure_warning,
        )

    def login_as(
        self,
        user: Union[User, str],
        password: str,
        email: str,
        force_sm_off: bool = False,
    ) -> LoginResponse:
        if isinstance(user, User):
            user = user.value

        response = self._api.login(user, email, password, force_sm_off)
        # The api returns an auth cookie which is automatically set
        # so there is no need to further process the response

        return LoginResponse.from_dict(response)

    def login(
        self, password: str, email: str = "", force_sm_off: bool = False
    ) -> LoginResponse:
        return self.login_as(User.CUSTOMER, password, email, force_sm_off)

    def logout(self) -> None:
        self._api.logout()

    def is_authenticated(self) -> bool:
        return self._api.is_authenticated()

    def run(self) -> None:
        self._api.get_sitemaster_run()

    def stop(self) -> None:
        self._api.get_sitemaster_stop()

    def get_charge(self) -> Union[float, int]:
        return assert_attribute(self._api.get_system_status_soe(), "percentage", "soe")

    def get_energy(self) -> int:
        return assert_attribute(
            self._api.get_system_status(),
            "nominal_energy_remaining",
            "system_status",
        )

    def get_sitemaster(self) -> SiteMasterResponse:
        return SiteMasterResponse.from_dict(self._api.get_sitemaster())

    def get_meters(self) -> MetersAggregatesResponse:
        return MetersAggregatesResponse.from_dict(self._api.get_meters_aggregates())

    def get_meter_site(self) -> MeterDetailsResponse:
        meter_response = self._api.get_meters_site()
        if meter_response is None or len(meter_response) == 0:
            raise ApiError("The powerwall returned no values for the site meter")

        return MeterDetailsResponse.from_dict(meter_response[0])

    def get_meter_solar(self) -> MeterDetailsResponse:
        meter_response = self._api.get_meters_solar()
        if meter_response is None or len(meter_response) == 0:
            raise ApiError("The powerwall returned no values for the solar meter")

        return MeterDetailsResponse.from_dict(meter_response[0])

    def get_grid_status(self) -> GridStatus:
        """Returns the current grid status."""
        status = assert_attribute(
            self._api.get_system_status_grid_status(),
            "grid_status",
            "grid_status",
        )

        return GridStatus(status)

    def get_capacity(self) -> float:
        return assert_attribute(
            self._api.get_system_status(),
            "nominal_full_pack_energy",
            "system_status",
        )

    def get_batteries(self) -> List[BatteryResponse]:
        batteries = assert_attribute(
            self._api.get_system_status(), "battery_blocks", "system_status"
        )
        return [BatteryResponse.from_dict(battery) for battery in batteries]

    def is_grid_services_active(self) -> bool:
        return assert_attribute(
            self._api.get_system_status_grid_status(),
            "grid_services_active",
            "grid_status",
        )

    def get_site_info(self) -> SiteInfoResponse:
        """Returns information about the powerwall site"""
        return SiteInfoResponse.from_dict(self._api.get_site_info())

    def set_site_name(self, site_name: str) -> dict:
        return self._api.post_site_info_site_name({"site_name": site_name})

    def get_status(self) -> PowerwallStatusResponse:
        return PowerwallStatusResponse.from_dict(self._api.get_status())

    def get_device_type(self) -> DeviceType:
        """Returns the device type of the powerwall"""
        return self.get_status().device_type

    def get_serial_numbers(self) -> List[str]:
        powerwalls = assert_attribute(
            self._api.get_powerwalls(), "powerwalls", "powerwalls"
        )
        return [
            assert_attribute(powerwall, "PackageSerialNumber")
            for powerwall in powerwalls
        ]

    def get_gateway_din(self) -> str:
        """Return the gateway din."""
        return assert_attribute(self._api.get_powerwalls(), "gateway_din", "powerwalls")

    def get_operation_mode(self) -> OperationMode:
        operation_mode = assert_attribute(
            self._api.get_operation(), "real_mode", "operation"
        )
        return OperationMode(operation_mode)

    def get_backup_reserve_percentage(self) -> float:
        return assert_attribute(
            self._api.get_operation(), "backup_reserve_percent", "operation"
        )

    def get_solars(self) -> List[SolarResponse]:
        return [SolarResponse.from_dict(solar) for solar in self._api.get_solars()]

    def get_vin(self) -> str:
        return assert_attribute(self._api.get_config(), "vin", "config")

    def set_island_mode(self, mode: IslandMode) -> IslandMode:
        return IslandMode(
            assert_attribute(
                self._api.post_islanding_mode({"island_mode": mode.value}),
                "island_mode",
            )
        )

    def get_version(self) -> str:
        version_str = assert_attribute(self._api.get_status(), "version", "status")
        return version_str.split(" ")[
            0
        ]  # newer versions include a sha trailer '21.44.1 c58c2df3'

    def get_api(self) -> API:
        return self._api

    def close(self) -> None:
        self._api.close()



# ruff: noqa: F401

from .api import API
from .const import (
    SUPPORTED_OPERATION_MODES,
    DeviceType,
    GridState,
    GridStatus,
    IslandMode,
    LineStatus,
    MeterType,
    OperationMode,
    Roles,
    SyncType,
    User,
)
from .error import (
    AccessDeniedError,
    ApiError,
    MeterNotAvailableError,
    MissingAttributeError,
    PowerwallError,
    PowerwallUnreachableError,
)
from .helpers import assert_attribute, convert_to_kw
from .powerwall import Powerwall
from .responses import (
    BatteryResponse,
    LoginResponse,
    MeterDetailsReadings,
    MeterDetailsResponse,
    MeterResponse,
    MetersAggregatesResponse,
    PowerwallStatusResponse,
    SiteInfoResponse,
    SiteMasterResponse,
    SolarResponse,
)

VERSION = "0.4.0"

__all__ = list(filter(lambda n: not n.startswith("_"), globals().keys()))



import re
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

from .const import DEFAULT_KW_ROUND_PERSICION, DeviceType, MeterType, Roles
from .error import MeterNotAvailableError
from .helpers import convert_to_kw


@dataclass
class ResponseBase:
    _raw: dict

    def __repr__(self) -> str:
        return str(self._raw)


@dataclass
class MeterResponse(ResponseBase):
    meter: MeterType
    instant_power: float
    last_communication_time: str
    frequency: float
    energy_exported: float
    energy_imported: float
    instant_total_current: float
    instant_average_voltage: float

    @staticmethod
    def from_dict(meter: MeterType, src: dict) -> "MeterResponse":
        return MeterResponse(
            src,
            meter=meter,
            instant_power=src["instant_power"],
            last_communication_time=src["last_communication_time"],
            frequency=src["frequency"],
            energy_exported=src["energy_exported"],
            energy_imported=src["energy_imported"],
            instant_total_current=src["instant_total_current"],
            instant_average_voltage=src["instant_average_voltage"],
        )

    def get_energy_exported(self, precision=DEFAULT_KW_ROUND_PERSICION) -> float:
        return convert_to_kw(self.energy_exported, precision)

    def get_energy_imported(self, precision=DEFAULT_KW_ROUND_PERSICION) -> float:
        return convert_to_kw(self.energy_imported, precision)

    def get_instant_total_current(self, precision=DEFAULT_KW_ROUND_PERSICION) -> float:
        return round(self.instant_total_current, precision)

    def get_power(self, precision=DEFAULT_KW_ROUND_PERSICION) -> float:
        return convert_to_kw(self.instant_power, precision)

    def is_active(self, precision=DEFAULT_KW_ROUND_PERSICION) -> bool:
        return self.get_power(precision) != 0

    def is_drawing_from(self, precision=DEFAULT_KW_ROUND_PERSICION) -> bool:
        if self.meter == MeterType.LOAD:
            # Cannot draw from load
            return False
        else:
            return self.get_power(precision) > 0

    def is_sending_to(self, precision=DEFAULT_KW_ROUND_PERSICION) -> bool:
        if self.meter == MeterType.LOAD:
            # For load the power is always positiv
            return self.get_power(precision) > 0
        else:
            return self.get_power(precision) < 0


@dataclass
class MeterDetailsReadings(MeterResponse):
    real_power_a: Optional[float]
    real_power_b: Optional[float]
    real_power_c: Optional[float]

    i_a_current: Optional[float]
    i_b_current: Optional[float]
    i_c_current: Optional[float]

    v_l1n: Optional[float]
    v_l2n: Optional[float]
    v_l3n: Optional[float]

    @staticmethod
    def from_dict(meter: MeterType, src: dict) -> "MeterDetailsReadings":
        meter_response = MeterResponse.from_dict(meter, src)
        return MeterDetailsReadings(
            real_power_a=src.get("real_power_a"),
            real_power_b=src.get("real_power_b"),
            real_power_c=src.get("real_power_c"),
            i_a_current=src.get("i_a_current"),
            i_b_current=src.get("i_b_current"),
            i_c_current=src.get("i_c_current"),
            v_l1n=src.get("v_l1n"),
            v_l2n=src.get("v_l2n"),
            v_l3n=src.get("v_l3n"),
            # Populate with the values from the base class
            **meter_response.__dict__
        )


@dataclass
class MeterDetailsResponse(ResponseBase):
    location: MeterType
    readings: MeterDetailsReadings

    @staticmethod
    def from_dict(src: dict) -> "MeterDetailsResponse":
        location = MeterType(src["location"])
        readings = MeterDetailsReadings.from_dict(location, src["Cached_readings"])
        return MeterDetailsResponse(src, location=location, readings=readings)


class MetersAggregatesResponse(ResponseBase):
    @staticmethod
    def from_dict(src: dict) -> "MetersAggregatesResponse":
        meters = {
            MeterType(key): MeterResponse.from_dict(MeterType(key), value)
            for key, value in src.items()
        }
        return MetersAggregatesResponse(src, meters)

    def __init__(self, response: dict, meters: Dict[MeterType, MeterResponse]) -> None:
        self._raw = response
        self.meters = meters

    def __getattribute__(self, attr) -> Any:
        if attr.upper() in MeterType.__dict__:
            m = MeterType(attr)
            if m in self.meters:
                return self.meters[m]
            else:
                raise MeterNotAvailableError(m, list(self.meters.keys()))
        else:
            return object.__getattribute__(self, attr)

    def get_meter(self, meter: MeterType) -> Optional[MeterResponse]:
        return self.meters.get(meter)


@dataclass
class SiteMasterResponse(ResponseBase):
    status: str
    is_running: bool
    is_connected_to_tesla: bool
    is_power_supply_mode: bool

    @staticmethod
    def from_dict(src: dict) -> "SiteMasterResponse":
        return SiteMasterResponse(
            src,
            status=src["status"],
            is_running=src["running"],
            is_connected_to_tesla=src["connected_to_tesla"],
            is_power_supply_mode=src["power_supply_mode"],
        )


@dataclass
class SiteInfoResponse(ResponseBase):
    nominal_system_energy: int
    nominal_system_power: int
    site_name: str
    timezone: str

    @staticmethod
    def from_dict(src: dict) -> "SiteInfoResponse":
        return SiteInfoResponse(
            src,
            nominal_system_energy=src["nominal_system_energy_kWh"],
            nominal_system_power=src["nominal_system_power_kW"],
            site_name=src["site_name"],
            timezone=src["timezone"],
        )


@dataclass
class PowerwallStatusResponse(ResponseBase):
    start_time: datetime
    up_time_seconds: timedelta
    version: str
    device_type: DeviceType
    commission_count: int
    sync_type: str
    git_hash: str

    _START_TIME_FORMAT = "%Y-%m-%d %H:%M:%S %z"
    _UP_TIME_SECONDS_REGEX = re.compile(
        r"^((?P<days>[\.\d]+?)d)?((?P<hours>[\.\d]+?)h)?((?P<minutes>[\.\d]+?)m)?((?P<seconds>[\.\d]+?)s)?$"
    )

    @staticmethod
    def _parse_uptime_seconds(up_time_seconds: str) -> timedelta:
        match = PowerwallStatusResponse._UP_TIME_SECONDS_REGEX.match(up_time_seconds)
        if not match:
            raise ValueError(
                "Unable to parse up time seconds {}".format(up_time_seconds)
            )

        time_params = {}
        for name, param in match.groupdict().items():
            if param:
                time_params[name] = float(param)

        return timedelta(**time_params)

    @staticmethod
    def from_dict(src: dict) -> "PowerwallStatusResponse":
        start_time = datetime.strptime(
            src["start_time"], PowerwallStatusResponse._START_TIME_FORMAT
        )
        up_time_seconds = PowerwallStatusResponse._parse_uptime_seconds(
            src["up_time_seconds"]
        )
        return PowerwallStatusResponse(
            src,
            start_time=start_time,
            up_time_seconds=up_time_seconds,
            version=src["version"],
            device_type=DeviceType(src["device_type"]),
            commission_count=src["commission_count"],
            sync_type=src["sync_type"],
            git_hash=src["git_hash"],
        )


@dataclass
class LoginResponse(ResponseBase):
    firstname: str
    lastname: str
    token: str
    roles: List[Roles]
    login_time: str

    @staticmethod
    def from_dict(src: dict) -> "LoginResponse":
        return LoginResponse(
            src,
            firstname=src["firstname"],
            lastname=src["lastname"],
            token=src["token"],
            roles=[Roles(role) for role in src["roles"]],
            login_time=src["loginTime"]
        )


@dataclass
class SolarResponse(ResponseBase):
    brand: str
    model: str
    power_rating_watts: int

    @staticmethod
    def from_dict(src: dict) -> "SolarResponse":
        return SolarResponse(
            src,
            brand=src["brand"],
            model=src["model"],
            power_rating_watts=src["power_rating_watts"],
        )


@dataclass
class BatteryResponse(ResponseBase):
    part_number: str
    serial_number: str
    energy_charged: int
    energy_discharged: int
    energy_remaining: int
    capacity: int
    wobble_detected: bool

    @staticmethod
    def from_dict(src: dict) -> "BatteryResponse":
        return BatteryResponse(
            src,
            part_number=src["PackagePartNumber"],
            serial_number=src["PackageSerialNumber"],
            energy_charged=src["energy_charged"],
            energy_discharged=src["energy_discharged"],
            energy_remaining=src["nominal_energy_remaining"],
            capacity=src["nominal_full_pack_energy"],
            wobble_detected=src["wobble_detected"],
        )



from typing import Union

from .error import MissingAttributeError


def convert_to_kw(value: float, precision: int = 1) -> float:
    """Converts watt to kilowatt and rounds to precision"""
    # Don't round if precision is -1
    if precision == -1:
        return value / 1000
    else:
        return round(value / 1000, precision)


def assert_attribute(response: dict, attr: str, url: Union[str, None] = None):
    value = response.get(attr)
    if value is None:
        raise MissingAttributeError(response, attr, url)
    else:
        return value
